MODULE domvvl
   !!======================================================================
   !!                       ***  MODULE domvvl   ***
   !! Ocean : 
   !!======================================================================
   !! History :   9.0  !  2006-06  (B. Levier, L. Marie)  original code
   !!              "   !  2007-07  (D. Storkey) add sigma option in vvl
   !!             3.0  !  2008-05  (M. Leclair, G. Madec)  bug corrections
   !!----------------------------------------------------------------------

   !!----------------------------------------------------------------------
   !!   dom_vvl         : defined scale factors & depths at each time step
   !!   dom_vvl_ini     : defined coefficients to distribute ssh on each layers
   !!   dom_vvl_ssh     : defined the ocean sea level at each time step
   !!----------------------------------------------------------------------
   USE oce             ! ocean dynamics and tracers
   USE dom_oce         ! ocean space and time domain
   USE sbc_oce         ! surface boundary condition: ocean
   USE dynspg_oce      ! surface pressure gradient variables
   USE phycst          ! physical constants
   USE in_out_manager  ! I/O manager
   USE lib_mpp         ! distributed memory computing library
   USE lbclnk          ! ocean lateral boundary conditions (or mpp link)
   USE bdy_oce         ! unstructured open boundary conditions

   IMPLICIT NONE
   PRIVATE

   PUBLIC dom_vvl_ini    ! called by dom_init.F90
   PUBLIC dom_vvl_ssh    ! called by trazdf.F90
   PUBLIC dom_vvl        ! called by istate.F90 and step.F90
   PUBLIC dom_vvl_sf_ini ! 
   PUBLIC dom_vvl_sf     ! 

   REAL(wp), PUBLIC, DIMENSION(jpi,jpj) ::   ee_t, ee_u, ee_v, ee_f   !:

   REAL(wp), DIMENSION(jpk) ::   r2dt   ! vertical profile of time-step, =2*rdttra except at nit000 (=rdttra) if neuler=0

   !! * Substitutions
#  include "domzgr_substitute.h90"
#  include "vectopt_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/OPA 3.0 , LOCEAN-IPSL (2008) 
   !! $Id: domvvl.F90 911 2008-04-28 09:31:32Z ctlod $
   !! Software governed by the CeCILL licence (modipsl/doc/NEMO_CeCILL.txt)
   !!----------------------------------------------------------------------

CONTAINS       

#if defined key_vvl
   !!----------------------------------------------------------------------
   !!   'key_vvl'                              variable volume
   !!----------------------------------------------------------------------

   SUBROUTINE dom_vvl_ini
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_ini  ***
      !!                   
      !! ** Purpose :   the coefficient used for the vertical redistribution
      !!              of the change in the ocean thickness due to the ssh
      !!                Here a sigma like redistribution is used, i.e. the
      !!              coefficients are the inverse of the local ocean depth
      !!              of the reference 3D levels.
      !!  
      !! ** Action  :   define ee_t, ee_u, ee_v and ee_f
      !!
      !!----------------------------------------------------------------------
      INTEGER  ::   ji, jj, jk
      !!----------------------------------------------------------------------

      IF(lwp)   THEN
         WRITE(numout,*)
         WRITE(numout,*) 'dom_vvl_ini : Variable volume activated'
         WRITE(numout,*) '~~~~~~~~~~~   compute coef. used to spread ssh over each layers'
      ENDIF

      IF( ln_zps )  CALL ctl_stop( 'dom_vvl_ini : option ln_zps is incompatible with variable volume option key_vvl')

#if defined key_zco  ||  defined key_dynspg_rl
      CALL ctl_stop( 'dom_vvl_ini : options key_zco/key_dynspg_rl are incompatible with variable volume option key_vvl')
#endif

      ! set fsvde.. and fsve3.. arrays to the reference fixed 3D level and level thickness
      fsvdept (:,:,:) = gdept (:,:,:)
      fsvdepw (:,:,:) = gdepw (:,:,:)
      fsvde3w (:,:,:) = gdep3w(:,:,:)
      fsve3t  (:,:,:) = e3t   (:,:,:)
      fsve3u  (:,:,:) = e3u   (:,:,:)
      fsve3v  (:,:,:) = e3v   (:,:,:)
      fsve3f  (:,:,:) = e3f   (:,:,:)
      fsve3w  (:,:,:) = e3w   (:,:,:)
      fsve3uw (:,:,:) = e3uw  (:,:,:)
      fsve3vw (:,:,:) = e3vw  (:,:,:)

      ! define ee_t, u, v and f as in sigma coordinate (ee_t = 1/ht, ...)
      ee_t(:,:) = fsve3t(ji,jj,1)        ! Lower bound : thickness of the first model level
      ee_u(:,:) = fsve3u(ji,jj,1)
      ee_v(:,:) = fsve3v(ji,jj,1)
      ee_f(:,:) = fsve3f(ji,jj,1)
      DO jk = 2, jpkm1                   ! Sum of the masked vertical scale factors
         ee_t(:,:) = ee_t(:,:) + fsve3t(ji,jj,jk) * tmask(ji,jj,jk)
         ee_u(:,:) = ee_u(:,:) + fsve3u(ji,jj,jk) * umask(ji,jj,jk)
         ee_v(:,:) = ee_v(:,:) + fsve3v(ji,jj,jk) * vmask(ji,jj,jk)
         DO jj = 1, jpjm1                      ! f-point case fmask cannot be used (fmask=shlat at solid boundaries)
            DO ji = 1, jpi                     ! take the product of umask instead
               ee_f(ji,jj) = ee_f(ji,jj) + fsve3f(ji,jj,jk) *  umask(ji,jj,jk) * umask(ji,jj+1,jk)
            END DO
         END DO
      END DO  
      !                                  ! Compute and mask the inverse of the local depth at T, U, V and F points
      ee_t(:,:) = 1. / ee_t(:,:) * tmask(:,:,1)
      ee_u(:,:) = 1. / ee_u(:,:) * umask(:,:,1)
      ee_v(:,:) = 1. / ee_v(:,:) * vmask(:,:,1)
      DO jj = 1, jpjm1                         ! f-point case fmask cannot be used 
         ee_f(:,jj) = 1. / ee_f(:,jj) * umask(:,jj,1) * umask(:,jj+1,1)
      END DO
      CALL lbc_lnk( ee_f, 'F', 1. )       ! lateral boundary condition on ee_f
      !
   END SUBROUTINE dom_vvl_ini


   SUBROUTINE dom_vvl
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl  ***
      !!                   
      !! ** Purpose :  compute ssh at U-V-F points, T-W scale factors and local
      !!               depths at each time step.
      !!----------------------------------------------------------------------
      INTEGER  ::   ji, jj, jk                 ! dummy loop indices
      REAL(wp), DIMENSION(jpi,jpj) :: zsshf    ! 2D workspace
      !!----------------------------------------------------------------------

      ! Sea Surface Height at u-v-fpoints
      DO jj = 1, jpjm1
         DO ji = 1,jpim1
            sshu(ji,jj) = 0.5 * umask(ji,jj,1) / ( e1u(ji,jj) * e2u(ji,jj) )  &
               &          * ( e1t(ji  ,jj) * e2t(ji  ,jj) * sshn(ji  ,jj)     &
               &            + e1t(ji+1,jj) * e2t(ji+1,jj) * sshn(ji+1,jj) )
            !
            sshv(ji,jj) = 0.5 * vmask(ji,jj,1) / ( e1v(ji,jj) * e2v(ji,jj) )  &
               &          * ( e1t(ji,jj  ) * e2t(ji,jj  ) * sshn(ji,jj  )     &
               &            + e1t(ji,jj+1) * e2t(ji,jj+1) * sshn(ji,jj+1) )
            !
            zsshf(ji,jj) = 0.25 * fmask(ji,jj,1)                 &
               &           * ( sshn(ji  ,jj) + sshn(ji  ,jj+1)   &
               &             + sshn(ji+1,jj) + sshn(ji+1,jj+1) )
         END DO
      END DO

      ! Boundaries conditions
      CALL lbc_lnk( sshu,  'U', 1. )
      CALL lbc_lnk( sshv,  'V', 1. )
      CALL lbc_lnk( zsshf, 'F', 1. )

      ! Scale factors at T levels
      DO jk = 1, jpkm1
         fse3t(:,:,jk) = fsve3t(:,:,jk) * ( 1 + sshn(:,:)  * ee_t(:,:) )
         fse3u(:,:,jk) = fsve3u(:,:,jk) * ( 1 + sshu(:,:)  * ee_u(:,:) )
         fse3v(:,:,jk) = fsve3v(:,:,jk) * ( 1 + sshv(:,:)  * ee_v(:,:) )
         fse3f(:,:,jk) = fsve3f(:,:,jk) * ( 1 + zsshf(:,:) * ee_f(:,:) )
      END DO

      ! Scale factors at W levels
      fse3w (:,:,1) = fse3t(:,:,1)
      fse3uw(:,:,1) = fse3u(:,:,1)
      fse3vw(:,:,1) = fse3v(:,:,1)
      DO jk = 2, jpk
         fse3w (:,:,jk) = 0.5 * ( fse3t(:,:,jk-1) + fse3t(:,:,jk) )
         fse3uw(:,:,jk) = 0.5 * ( fse3u(:,:,jk-1) + fse3u(:,:,jk) )
         fse3vw(:,:,jk) = 0.5 * ( fse3v(:,:,jk-1) + fse3v(:,:,jk) )
      END DO

      ! T and W points depth
      fsdept(:,:,1) = 0.5 * fse3w(:,:,1)
      fsdepw(:,:,1) = 0.e0
      fsde3w(:,:,1) = fsdept(:,:,1) - sshn(:,:)
      DO jk = 2, jpk
         fsdept(:,:,jk) = fsdept(:,:,jk-1) + fse3w(:,:,jk)
         fsdepw(:,:,jk) = fsdepw(:,:,jk-1) + fse3t(:,:,jk-1)
         fsde3w(:,:,jk) = fsdept(:,:,jk  ) - sshn(:,:)
      END DO

      IF( MINVAL(fse3t(:,:,:)) < 0.0 ) THEN
         CALL ctl_stop('E R R O R : dom_vvl','Level thickness fse3t less than zero.')
         nstop = nstop+1
      ENDIF

      ! Local depth or Inverse of the local depth of the water column at u- and v-points
      ! ------------------------------

      ! Ocean depth at U- and V-points
      hu(:,:) = 0.e0    ;    hv(:,:) = 0.e0

      DO jk = 1, jpk
         hu(:,:) = hu(:,:) + fse3u(:,:,jk) * umask(:,:,jk)
         hv(:,:) = hv(:,:) + fse3v(:,:,jk) * vmask(:,:,jk)
      END DO


      ! Inverse of the local depth
      hur(:,:) = fse3u(:,:,1)             ! Lower bound : thickness of the first model level
      hvr(:,:) = fse3v(:,:,1)
      
      DO jk = 2, jpk                      ! Sum of the vertical scale factors
         hur(:,:) = hur(:,:) + fse3u(:,:,jk) * umask(:,:,jk)
         hvr(:,:) = hvr(:,:) + fse3v(:,:,jk) * vmask(:,:,jk)
      END DO

      ! Compute and mask the inverse of the local depth
      hur(:,:) = 1. / hur(:,:) * umask(:,:,1)
      hvr(:,:) = 1. / hvr(:,:) * vmask(:,:,1)

   END SUBROUTINE dom_vvl


   SUBROUTINE dom_vvl_ssh( kt ) 
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_ssh  ***
      !!                   
      !! ** Purpose :  compute the ssha field used in tra_zdf, dynnxt, tranxt 
      !!               and dynspg routines
      !!----------------------------------------------------------------------
      !! * Arguments
      INTEGER, INTENT( in )  ::    kt                         ! time step

      !! * Local declarations
      INTEGER  ::   ji, jj, jk                                ! dummy loop indices
      REAL(wp) ::   z2dt, zraur                               ! temporary scalars
      REAL(wp), DIMENSION(jpi,jpj) ::   zun, zvn, zhdiv       ! 2D workspace
      !!----------------------------------------------------------------------

      !! Sea surface height after (ssha) in variable volume case
      !! --------------------------====-----===============-----
      !! ssha is needed for the tracer time-stepping (trazdf_imp or
      !! tra_nxt), for the ssh time-stepping (dynspg_*) and for the
      !! dynamics time-stepping (dynspg_flt or dyn_nxt, and wzv).
      !! un, vn (or un_b and vn_b) and emp are needed, so it must be
      !! done after the call of oce_sbc.

      IF( neuler == 0 .AND. kt == nit000 ) THEN     ! at nit000
         r2dt(:) =  rdttra(:)                       ! = rdtra (restarting with Euler time stepping)
      ELSEIF( kt <= nit000 + 1) THEN                ! at nit000 or nit000+1
         r2dt(:) = 2. * rdttra(:)                   ! = 2 rdttra (leapfrog)
      ENDIF

      z2dt  = r2dt(1)
      zraur = 1. / rauw

      ! Vertically integrated quantities
      ! --------------------------------
      IF( .NOT. lk_dynspg_ts .OR. (neuler == 0 .AND. kt == nit000) ) THEN
         zun(:,:) = 0.e0    ;    zvn(:,:) = 0.e0
         !
         DO jk = 1, jpkm1
            !                                               ! Vertically integrated transports (now)
            zun(:,:) = zun(:,:) + fse3u(:,:,jk) * un(:,:,jk)
            zvn(:,:) = zvn(:,:) + fse3v(:,:,jk) * vn(:,:,jk)
         END DO
      ELSE ! lk_dynspg_ts is true
         zun (:,:) = un_b(:,:)    ;    zvn (:,:) = vn_b(:,:)
      ENDIF

      ! Horizontal divergence of barotropic transports
      !--------------------------------------------------
      DO jj = 2, jpjm1
         DO ji = 2, jpim1   ! vector opt.
            zhdiv(ji,jj) = (  e2u(ji  ,jj  ) * zun(ji  ,jj  )    &
               &            - e2u(ji-1,jj  ) * zun(ji-1,jj  )    &
               &            + e1v(ji  ,jj  ) * zvn(ji  ,jj  )    &
               &            - e1v(ji  ,jj-1) * zvn(ji  ,jj-1) )  &
               &           / ( e1t(ji,jj) * e2t(ji,jj) )
         END DO
      END DO

#if defined key_obc && ( key_dynspg_exp || key_dynspg_ts )
      ! open boundaries (div must be zero behind the open boundary)
      !  mpp remark: The zeroing of hdiv can probably be extended to 1->jpi/jpj for the correct row/column
      IF( lp_obc_east  )   zhdiv(nie0p1:nie1p1,nje0  :nje1)   = 0.e0    ! east
      IF( lp_obc_west  )   zhdiv(niw0  :niw1  ,njw0  :njw1)   = 0.e0    ! west
      IF( lp_obc_north )   zhdiv(nin0  :nin1  ,njn0p1:njn1p1) = 0.e0    ! north
      IF( lp_obc_south )   zhdiv(nis0  :nis1  ,njs0  :njs1)   = 0.e0    ! south
#endif

#if defined key_bdy || key_bdy_tides
      DO jj = 1, jpj
         DO ji = 1, jpi
            zhdiv(ji,jj) = zhdiv(ji,jj)*bdytmask(ji,jj)
         END DO
      END DO
#endif

      CALL lbc_lnk( zhdiv, 'T', 1. )

      ! Sea surface elevation time stepping
      ! -----------------------------------
      IF( .NOT. lk_dynspg_ts .OR. (neuler == 0 .AND. kt == nit000) ) THEN
         ssha(:,:) = sshb(:,:) - z2dt * ( zraur * emp(:,:) + zhdiv(:,:) ) * tmask(:,:,1)
      ELSE ! lk_dynspg_ts is true
         ssha(:,:) = (  sshb_b(:,:) - z2dt * ( zraur * emp(:,:) + zhdiv(:,:) )  ) * tmask(:,:,1)
      ENDIF


   END SUBROUTINE dom_vvl_ssh

   SUBROUTINE  dom_vvl_sf( zssh, gridp, sfe3 )
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_sf  ***
      !!                   
      !! ** Purpose :  compute vertical scale factor at each time step
      !!----------------------------------------------------------------------
      !! * Arguments
      CHARACTER(LEN=1)                , INTENT( in ) :: gridp   ! grid point type
      REAL(wp), DIMENSION(jpi,jpj)    , INTENT( in ) :: zssh    ! 2D workspace
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT( out) :: sfe3    ! 3D workspace

      !! * Local declarations
      INTEGER  ::   jk                                      ! dummy loop indices
      !!----------------------------------------------------------------------

      SELECT CASE (gridp)

      CASE ('U')
         !
         DO jk = 1, jpk
            sfe3(:,:,jk)  = fsve3u(:,:,jk) * ( 1 + zssh(:,:) * ee_u(:,:) )
         ENDDO

      CASE ('V')
         !
         DO jk = 1, jpk
            sfe3(:,:,jk)  = fsve3v(:,:,jk) * ( 1 + zssh(:,:) * ee_v(:,:) )
         ENDDO

      CASE ('T')
         !
         DO jk = 1, jpk
            sfe3(:,:,jk)  = fsve3t(:,:,jk) * ( 1 + zssh(:,:) * ee_t(:,:) )
         ENDDO

      END SELECT

   END SUBROUTINE dom_vvl_sf

   SUBROUTINE dom_vvl_sf_ini( gridp, sfe3ini )
      !!----------------------------------------------------------------------
      !!                ***  ROUTINE dom_vvl_sf_ini  ***
      !!                   
      !! ** Purpose :  affect the appropriate vertical scale factor. It is done
      !!               to avoid compiling error when using key_zco cpp key
      !!----------------------------------------------------------------------
      !! * Arguments
      CHARACTER(LEN=1)                , INTENT( in ) :: gridp      ! grid point type
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT (out) :: sfe3ini    ! 3D workspace
      !!----------------------------------------------------------------------

      SELECT CASE (gridp)

      CASE ('U')
         !
         sfe3ini(:,:,:)  = fse3u(:,:,:)

      CASE ('V')
         !
         sfe3ini(:,:,:)  = fse3v(:,:,:)

      CASE ('T')
         !
         sfe3ini(:,:,:)  = fse3t(:,:,:)

      END SELECT

   END SUBROUTINE dom_vvl_sf_ini
#else
   !!----------------------------------------------------------------------
   !!   Default option :                                      Empty routine
   !!----------------------------------------------------------------------
   SUBROUTINE dom_vvl_ini
   END SUBROUTINE dom_vvl_ini
   SUBROUTINE dom_vvl
   END SUBROUTINE dom_vvl
   SUBROUTINE dom_vvl_ssh( kt ) 
      !! * Arguments
      INTEGER, INTENT( in )  ::    kt        ! time step
      WRITE(*,*) 'dom_vvl_ssh: You should not have seen this print! error?', kt
   END SUBROUTINE dom_vvl_ssh
   SUBROUTINE dom_vvl_sf( zssh, gridp, sfe3 ) 
      !! * Arguments
      CHARACTER(LEN=1)                , INTENT( in ) :: gridp   ! grid point type
      REAL(wp), DIMENSION(jpi,jpj)    , INTENT( in ) :: zssh    ! 2D workspace
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT (out) :: sfe3    ! 3D workspace
      sfe3(:,:,:) = 0.e0
      WRITE(*,*) 'sfe3: You should not have seen this print! error?', gridp
      WRITE(*,*) 'sfe3: You should not have seen this print! error?', zssh(1,1)
   END SUBROUTINE dom_vvl_sf
   SUBROUTINE dom_vvl_sf_ini( gridp, sfe3ini ) 
      !! * Arguments
      CHARACTER(LEN=1)                , INTENT( in ) :: gridp    ! grid point type
      REAL(wp), DIMENSION(jpi,jpj,jpk), INTENT (out) :: sfe3ini  ! 3D workspace
      sfe3ini(:,:,:) = 0.e0
      WRITE(*,*) 'sfe3ini: You should not have seen this print! error?', gridp
   END SUBROUTINE dom_vvl_sf_ini
#endif

   !!======================================================================
END MODULE domvvl